"""Managers for reviewboard.diffviewer.models."""

from __future__ import annotations

import bz2
import gc
import hashlib
import logging
from functools import partial
from typing import Mapping, Sequence, TYPE_CHECKING

from django.conf import settings
from django.db import models, reset_queries, connection, connections
from django.db.models import Count, Q
from django.db.utils import IntegrityError
from django.utils.translation import gettext as _

from reviewboard.diffviewer.commit_utils import get_file_exists_in_history
from reviewboard.diffviewer.differ import DiffCompatVersion
from reviewboard.diffviewer.diffutils import check_diff_size
from reviewboard.diffviewer.filediff_creator import create_filediffs

if TYPE_CHECKING:
    from reviewboard.diffviewer.models.diffcommit import DiffCommit


logger = logging.getLogger(__name__)


class FileDiffManager(models.Manager):
    """A manager for FileDiff objects.

    This contains utility methods for locating FileDiffs that haven't been
    migrated to use RawFileDiffData.
    """

    MIGRATE_OBJECT_LIMIT = 200

    def unmigrated(self):
        """Query FileDiffs that store their own diff content.

        This will return FileDiffs that were created prior to Review Board 1.7.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for filtering FileDiffs that have not been migrated to
            use some form of deduplicated diff storage mechanism.
        """
        return self.filter(
            Q(diff_hash_id__isnull=True) &
            Q(legacy_diff_hash_id__isnull=True) &
            Q(parent_diff_hash_id__isnull=True) &
            Q(legacy_parent_diff_hash_id__isnull=True))

    def get_migration_counts(self):
        """Return the number of items that need to be migrated.

        The result is a dictionary containing a breakdown of the various
        counts, and the total count of all items for display.

        Returns:
            dict:
            A dictionary of counts. This will contain the following keys:

            ``filediffs``:
                The number of unmigrated FileDiff objects.

            ``legacy_file_diff_data``:
                The number of unmigrated LegacyFileDiffData objects.

            ``total_count``:
                The total count of objects to migrate.
        """
        from reviewboard.diffviewer.models import LegacyFileDiffData

        unmigrated_filediffs_count = self.unmigrated().count()
        legacy_fdd_count = None
        warning = None

        legacy_fdd_cnx = connections[LegacyFileDiffData.objects.db]

        if legacy_fdd_cnx.vendor == 'mysql':
            # On MySQL, computing the number of LegacyFileDiffData objects
            # can be very slow when using InnoDB tables.
            #
            # Unlike MyISAM, which keeps a total row count for easy
            # reference, InnoDB requires walking the index (stored in a
            # BTree) to get the total row count, and for this table, that's
            # expensive.
            #
            # A large part of the reason why is because we made a design
            # error early on and used SHA1 values as the primary key. This
            # makes for a larger, expensive index, and MySQL seems to hate
            # this.
            #
            # So the workaround is that we're just going to ask MySQL for
            # its last-known row count. For MyISAM, this is going to be an
            # exact value. For InnoDB, it might be an estimate.
            #
            # The MySQL documentation says that the estimate may be off by
            # as much as 40%-50%. For this reason, we'll warn the user, but
            # it won't actually harm the migration process. The total count
            # is purely informative.
            #
            # Note that running `ANALYZE TABLE diffviewer_filediffdata`
            # before the migration will sync up the values, but that operation
            # may take some time.
            table_name = LegacyFileDiffData._meta.db_table

            try:
                cursor = legacy_fdd_cnx.cursor()
                cursor.execute('SHOW TABLE STATUS WHERE Name="%s"'
                               % table_name)
                result = cursor.fetchone()

                # We have to fetch these by index. These should be stable.
                table_type = result[1]
                legacy_fdd_count = result[4]

                if legacy_fdd_count == 0:
                    # This might be correct, but since 0 means "we're done,"
                    # we want to be sure. Let's force a query below.
                    legacy_fdd_count = None

                if table_type == 'InnoDB':
                    warning = _(
                        'The diff migration count is just an estimate. The '
                        '%s table is backed by InnoDB, which does not '
                        'provide up-to-date row counts, and querying can be '
                        'too slow for this table. This will only affect '
                        'progress notification and will not otherwise impact '
                        'diff migration.'
                    ) % table_name
            except Exception as e:
                # Something went wrong. We're going to fall back on
                # calculating from the database, though it will be slow.
                logger.exception('Unable to fetch information on the %s '
                                 'table: %s',
                                 table_name, e)
                logger.warning('Calculating the number of diffs in the '
                               '%s table. This may take a while...',
                               table_name)

        if legacy_fdd_count is None:
            legacy_fdd_count = LegacyFileDiffData.objects.count()

        return {
            'filediffs': unmigrated_filediffs_count,
            'legacy_file_diff_data': legacy_fdd_count,
            'total_count': unmigrated_filediffs_count + legacy_fdd_count,
            'warning': warning,
        }

    def migrate_all(self, batch_done_cb=None, counts=None, batch_size=40,
                    max_diffs=None):
        """Migrate diff content in FileDiffs to use RawFileDiffData.

        This will run through all unmigrated FileDiffs and migrate them,
        condensing their storage needs and removing the content from
        FileDiffs.

        This will return a dictionary with the result of the process.

        Args:
            batch_done_cb (callable, optional):
                A function to call after each batch of objects has been
                processed. This can be used for progress notification.

                This should be in the form of:

                .. code-block:: python

                   def on_batch_done(total_diffs_migrated=None,
                                     total_count=None, **kwargs):
                       ...

                Note that ``total_count`` may be ``None``.

            counts (dict, optional):
                A dictionary of counts for calculations.

                The only value used is ``total_count``, which would be
                a total number of objects being processed.

                This is only used for reporting to ``batch_done_cb``. If
                not provided, and ``batch_done_cb`` *is* provided, then
                this method will query the counts itself.

            batch_size (int, optional):
                The number of objects to process in each batch.

            max_diffs (int, optional):
                The maximum number of diffs to migrate.
        """
        from reviewboard.diffviewer.models import LegacyFileDiffData

        assert batch_done_cb is None or callable(batch_done_cb)

        total_diffs_migrated = 0
        total_diff_size = 0
        total_bytes_saved = 0

        unmigrated_filediffs = self.unmigrated()
        legacy_data_items = LegacyFileDiffData.objects.all()

        if counts is not None:
            total_count = counts.get('total_count')
        else:
            total_count = self.get_migration_counts()

        if max_diffs is not None:
            if total_count is None:
                total_count = max_diffs
            else:
                total_count = min(total_count, max_diffs)

        migration_tasks = (
            (self._migrate_filediffs, unmigrated_filediffs),
            (self._migrate_legacy_fdd, legacy_data_items),
        )

        for migrate_func, queryset in migration_tasks:
            for batch_info in migrate_func(queryset=queryset,
                                           batch_size=batch_size,
                                           max_diffs=max_diffs):
                total_diffs_migrated += batch_info[0]
                total_diff_size += batch_info[1]
                total_bytes_saved += batch_info[2]

                if batch_done_cb is not None:
                    batch_done_cb(total_diffs_migrated=total_diffs_migrated,
                                  total_count=total_count)

                if max_diffs is not None:
                    max_diffs -= batch_info[0]

            if max_diffs is not None and max_diffs <= 0:
                break

        # Call batch_done_cb one more time, using the finalized total count
        # which may differ from the original count due to a bad total row
        # count estimate or a too-large max_diffs.
        if batch_done_cb is not None:
            batch_done_cb(total_diffs_migrated=total_diffs_migrated,
                          total_count=total_diffs_migrated)

        return {
            'diffs_migrated': total_diffs_migrated,
            'old_diff_size': total_diff_size,
            'new_diff_size': total_diff_size - total_bytes_saved,
            'bytes_saved': total_bytes_saved,
        }

    def _migrate_legacy_fdd(self, queryset, batch_size,
                            max_diffs=None):
        """Migrate data from LegacyFileDiffData to RawFileDiffData.

        This will go through every
        :py:class:`~reviewboard.diffviewer.models.LegacyFileDiffData` and
        convert them to
        :py:class:`~reviewboard.diffviewer.models.RawFileDiffData` entries,
        removing the old versions. All associated FileDiffs are then updated to
        point to the new RawFileDiffData entry instead of the old
        LegacyFileDiffData.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset for retrieving
                :py:class:`~reviewboard.diffviewer.models.LegacyFileDiffData`
                objects.

            batch_size (int):
                The number of objects to process in each batch.

            max_diffs (int, optional):
                The maximum number of diffs to migrate. This may be ``None``,
                in which case all diffs will be migrated.

        Yields:
            tuple:
            A tuple containing the following items:

            1. The size of the batch.
            2. The total number of bytes of diff data from the old legacy
               entries in this batch.
            3. The total number of bytes saved during this migration.
            4. A list of all legacy hashes that were migrated for diffs.
            5. A list of all legacy hashes that were migrated for parent diffs.
            6. A list of all legacy hashes that were migrated for diffs and
               parent diffs.
        """
        from reviewboard.diffviewer.models import RawFileDiffData

        cursor = connection.cursor()

        queryset = queryset.annotate(
            num_filediffs=Count('filediffs'),
            num_parent_filediffs=Count('parent_filediffs'))

        for batch in self._iter_batches(queryset=queryset,
                                        batch_size=batch_size,
                                        max_diffs=max_diffs):
            batch_total_diff_size = 0
            batch_total_bytes_saved = 0
            raw_fdds = []
            all_diff_hashes = []
            filediff_hashes = []
            parent_filediff_hashes = []

            for legacy_fdd in batch:
                raw_fdd = RawFileDiffData.objects.create_from_legacy(
                    legacy_fdd, save=False)

                raw_fdds.append(raw_fdd)

                binary_hash = legacy_fdd.binary_hash

                old_diff_size = len(legacy_fdd.get_binary_base64())
                batch_total_diff_size += old_diff_size
                batch_total_bytes_saved += old_diff_size - len(raw_fdd.binary)

                # Update all associated FileDiffs to use the new objects
                # instead of the old ones.
                if legacy_fdd.num_filediffs > 0:
                    filediff_hashes.append(binary_hash)

                if legacy_fdd.num_parent_filediffs > 0:
                    parent_filediff_hashes.append(binary_hash)

                all_diff_hashes.append(binary_hash)

            try:
                # Attempt to create all the entries we want in one go.
                RawFileDiffData.objects.bulk_create(raw_fdds)
            except IntegrityError:
                # One or more entries in the batch conflicted with an existing
                # entry, meaning it was already created. We'll just need to
                # operate on the contents of this batch one-by-one.
                for raw_fdd in raw_fdds:
                    try:
                        raw_fdd.save()
                    except IntegrityError:
                        raw_fdd = RawFileDiffData.objects.get(
                            binary_hash=raw_fdd.binary_hash)

                        # This was already in the database, so we didn't have
                        # to write new data. That means we get to reclaim
                        # its size in the amount of bytes saved.
                        batch_total_bytes_saved += len(raw_fdd.binary)

            if filediff_hashes:
                self._transition_hashes(cursor, 'diff_hash', filediff_hashes)

            if parent_filediff_hashes:
                self._transition_hashes(cursor, 'parent_diff_hash',
                                        parent_filediff_hashes)

            queryset.filter(pk__in=all_diff_hashes).delete()

            yield (len(batch), batch_total_diff_size,
                   batch_total_bytes_saved, filediff_hashes,
                   parent_filediff_hashes, all_diff_hashes)

    def _migrate_filediffs(self, queryset, batch_size, max_diffs=None):
        """Migrate old diff data from a FileDiff into a RawFileDiffData.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset for retrieving
                :py:class:`~reviewboard.diffviewer.models.FileDiff` objects.

            batch_size (int):
                The number of objects to process in each batch.

            max_diffs (int, optional):
                The maximum number of diffs to migrate. This may be ``None``,
                in which case all diffs will be migrated.

        Yields:
            tuple:
            A tuple containing the following items:

            1. The size of the batch.
            2. The total number of bytes of diff data from the old legacy
               entries in this batch.
            3. The total number of bytes saved during this migration.
        """
        for batch in self._iter_batches(queryset=queryset,
                                        batch_size=batch_size,
                                        max_diffs=max_diffs):
            batch_total_diff_size = 0
            batch_total_bytes_saved = 0

            for filediff in batch:
                diff_size = len(filediff.get_diff64_base64())
                parent_diff_size = len(filediff.get_parent_diff64_base64())

                batch_total_diff_size += diff_size + parent_diff_size

                diff_hash_is_new, parent_diff_hash_is_new = \
                    filediff._migrate_diff_data(recalculate_counts=False)

                if diff_size > 0:
                    batch_total_bytes_saved += diff_size

                    if diff_hash_is_new:
                        # This is a new entry, so we have to subtract the
                        # new storage size. This *could* be larger than the
                        # original diff, but will usually be smaller.
                        batch_total_bytes_saved -= \
                            len(filediff.diff_hash.binary)

                if parent_diff_size > 0:
                    batch_total_bytes_saved += parent_diff_size

                    if diff_hash_is_new:
                        # This is a new entry, so we have to subtract the
                        # new storage size. This *could* be larger than the
                        # original diff, but will usually be smaller.
                        batch_total_bytes_saved -= \
                            len(filediff.parent_diff_hash.binary)

            yield len(batch), batch_total_diff_size, batch_total_bytes_saved

    def _iter_batches(self, queryset, batch_size, max_diffs=None,
                      object_limit=MIGRATE_OBJECT_LIMIT):
        """Iterate through items in a queryset, yielding batches.

        This will gather up to a specified number of items from a
        queryset at a time, process them into batches of a specified
        size, and yield them.

        After each set of objects fetched from the database, garbage
        collection will be forced and stored queries reset, in order to
        reduce memory usage.

        Args:
            queryset (django.db.models.query.QuerySet):
                The queryset to execute for fetching objects.

            batch_size (int):
                The maximum number of objects to yield per batch.

            max_diffs (int, optional):
                The maximum number of diffs to migrate. This may be ``None``,
                in which case all diffs will be migrated.

            object_limit (int, optional):
                The maximum number of objects to fetch from the database per
                query.

        Yields:
            list of django.db.models.Model:
            A batch of items to process. This will never be larger than
            ``batch_size``.
        """
        batch = []
        total_processed = 0

        # We're going to iterate until we've exhausted the available objects
        # from the database for the provided query, or hit the maximum number
        # of diffs that were requested.
        while max_diffs is None or total_processed < max_diffs:
            # Every time we work on a batch, we're re-querying the list of
            # objects. This result from the query is expected not to have any
            # previously-processed objects from a yielded batch. It may,
            # however, have objects we've previously seen that haven't been
            # yielded in a batch yet. That's why we're indexing from the
            # length of the batch to the object limit (or to the requested
            # max_diffs, whichever is smaller).
            limit = object_limit
            processed = 0

            if max_diffs is not None:
                limit = min(limit, max_diffs - total_processed)

            batch_len = len(batch)

            for item in queryset[batch_len:limit].iterator():
                batch.append(item)
                processed += 1

                if len(batch) == batch_size:
                    yield batch
                    batch = []

            total_processed += processed

            # Do all we can to limit the memory usage by resetting any
            # stored queries (if DEBUG is True), and force garbage
            # collection of anything we may have from processing an object.
            reset_queries()
            gc.collect()

            if processed < limit:
                # We've processed all items in the database. We're done.
                break

        if batch:
            yield batch

    def _transition_hashes(self, cursor, hash_field_name, diff_hashes):
        """Transitions FileDiff-associated hashes to RawFileDiffData.

        This queries all FileDiffs and RawFileDiffData entries referencing
        the given list of diff hashes, and updates the FileDiffs to point
        to those instead of the formerly-associated LegacyFileDiffDatas.
        """
        from reviewboard.diffviewer.models import RawFileDiffData

        legacy_hash_field_name = 'legacy_%s' % hash_field_name

        # Since this is a pretty complex operation, we're going to sanity-check
        # results on DEBUG setups, to help catch issues that might come up as
        # this code changes.
        if settings.DEBUG:
            old_filediff_info = dict(
                (filediff.pk, getattr(filediff, legacy_hash_field_name).pk)
                for filediff in self.filter(**{
                    legacy_hash_field_name + '__in': diff_hashes,
                })
            )
        else:
            old_filediff_info = None

        # If the database supports joins on updates, then we can craft
        # a query that will massively speed up the diff transition time.
        # Otherwise, we need to fall back on doing a select and then an
        # update per result.
        #
        # The queries are different between databases (yay standards), so
        # we can't be smart and do this in a generic way. We have to check
        # the database types.
        if connection.vendor == 'mysql':
            cursor.execute(
                'UPDATE %(filediff_table)s'
                '  INNER JOIN %(raw_fdd_table)s raw_fdd'
                '    ON raw_fdd.binary_hash = '
                '       %(filediff_table)s.%(hash_field_name)s_id'
                '  SET'
                '    raw_%(hash_field_name)s_id = raw_fdd.id,'
                '    %(hash_field_name)s_id = NULL'
                '  WHERE raw_fdd.binary_hash IN (%(diff_hashes)s)'
                % {
                    'filediff_table': self.model._meta.db_table,
                    'raw_fdd_table': RawFileDiffData._meta.db_table,
                    'hash_field_name': hash_field_name,
                    'diff_hashes': ','.join(
                        "'%s'" % diff_hash
                        for diff_hash in diff_hashes
                    ),
                })
        elif connection.vendor == 'postgresql':
            cursor.execute(
                'UPDATE %(filediff_table)s'
                '  SET'
                '    raw_%(hash_field_name)s_id = raw_fdd.id,'
                '    %(hash_field_name)s_id = NULL'
                '  FROM %(raw_fdd_table)s raw_fdd'
                '  WHERE'
                '    raw_fdd.binary_hash IN (%(diff_hashes)s) AND'
                '    raw_fdd.binary_hash = '
                '        %(hash_field_name)s_id'
                % {
                    'filediff_table': self.model._meta.db_table,
                    'raw_fdd_table': RawFileDiffData._meta.db_table,
                    'hash_field_name': hash_field_name,
                    'diff_hashes': ','.join(
                        "'%s'" % diff_hash
                        for diff_hash in diff_hashes
                    ),
                })
        else:
            raw_fdds = RawFileDiffData.objects.filter(
                binary_hash__in=diff_hashes).only('pk', 'binary_hash')

            for raw_fdd in raw_fdds:
                self.filter(**{
                    legacy_hash_field_name: raw_fdd.binary_hash
                }).update(**{
                    hash_field_name: raw_fdd.pk,
                    legacy_hash_field_name: None
                })

        if settings.DEBUG:
            new_filediff_info = dict(
                (filediff.pk, getattr(filediff, hash_field_name).binary_hash)
                for filediff in self.filter(pk__in=old_filediff_info.keys())
            )

            assert old_filediff_info == new_filediff_info


class RawFileDiffDataManager(models.Manager):
    """A custom manager for RawFileDiffData.

    This provides conveniences for creating an entry based on a
    LegacyFileDiffData object.
    """
    def process_diff_data(self, data):
        """Processes a diff, returning the resulting content and compression.

        If the content would benefit from being compressed, this will
        return the compressed content and the value for the compression
        flag. Otherwise, it will return the raw content.
        """
        compressed_data = bz2.compress(data, 9)

        if len(compressed_data) < len(data):
            return compressed_data, self.model.COMPRESSION_BZIP2
        else:
            return data, None

    def get_or_create_from_data(self, data):
        """Return or create a new stored entry for diff data.

        Args:
            data (bytes):
                The diff data to store or return an entry for.

        Returns:
            reviewboard.diffviewer.models.raw_file_diff_data.RawFileDiffData:
            The entry for the diff data.

        Raises:
            TypeError:
                The data passed in was not a bytes string.
        """
        if not isinstance(data, bytes):
            raise TypeError(
                'RawFileDiffData.objects.get_or_create_data expects a '
                'bytes value, not %s'
                % type(data))

        binary_hash = self._hash_hexdigest(data)
        processed_data, compression = self.process_diff_data(data)

        return self.get_or_create(
            binary_hash=binary_hash,
            defaults={
                'binary': processed_data,
                'compression': compression,
            })

    def create_from_legacy(self, legacy, save=True):
        processed_data, compression = self.process_diff_data(legacy.binary)

        raw_file_diff_data = self.model(binary_hash=legacy.binary_hash,
                                        binary=processed_data,
                                        compression=compression)
        raw_file_diff_data.extra_data = legacy.extra_data

        if save:
            raw_file_diff_data.save()

        return raw_file_diff_data

    def _hash_hexdigest(self, diff):
        hasher = hashlib.sha1()
        hasher.update(diff)
        return hasher.hexdigest()


class BaseDiffManager(models.Manager):
    """A base manager class for creating models out of uploaded diffs"""

    def create_from_data(self, *args, **kwargs):
        """Create a model instance from data.

        Subclasses must implement this method.

        See :py:meth:`create_from_upload` for the fields that will be passed to
        this method.

        Returns:
            django.db.models.Model:
            The model instance.
        """
        raise NotImplementedError

    def create_from_upload(self, repository, diff_file, parent_diff_file=None,
                           request=None, validate_only=False, **kwargs):
        """Create a model instance from a form upload.

        This parses a diff and optional parent diff covering one or more files,
        validates, and constructs either:

        * a :py:class:`~reviewboard.diffviewer.models.diffset.DiffSet` or
        * a :py:class:`~reviewboard.diffviewer.models.diffcommit.DiffCommit`

        and the child :py:class:`FileDiffs
        <reviewboard.diffviewer.models.filediff.FileDiff>` representing the
        diff.

        This can optionally validate the diff without saving anything to the
        database. In this case, no value will be returned. Instead, callers
        should take any result as success.

        This function also accepts a number of keyword arguments when creating
        a :py:class:`~reviewboard.diffviewer.model.DiffCommit`. In that case,
        the following fields are required (except for the committer fields when
        the underlying SCM does not distinguish between author and committer):

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the diff applies to.

            diff_file (django.core.files.uploadedfile.UploadedFile):
                The diff file uploaded in the form.

            parent_diff_file (django.core.files.uploadedfile.UploadedFile,
                              optional):
                The parent diff file uploaded in the form.

            request (django.http.HttpRequest, optional):
                The current HTTP request, if any. This will result in better
                logging.

            validate_only (bool, optional):
                Whether to just validate and not save. If ``True``, then this
                won't populate the database at all and will return ``None``
                upon success. This defaults to ``False``.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`create_from_data`.

                See :py:meth:`~DiffSetManager.create_from_data` and
                :py:meth:`DiffCommitManager.create_from_data` for acceptable
                keyword arguments.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The resulting DiffSet stored in the database, if processing
            succeeded and ``validate_only=False``.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the main diff or parent diff.

            reviewboard.diffviewer.errors.DiffTooBigError:
                The diff file was too big to be uploaded, based on the
                configured maximum diff size in settings.

            reviewboard.diffviewer.errors.EmptyDiffError:
                The provided diff file did not contain any file changes.

            reviewboard.scmtools.core.FileNotFoundError:
                A file specified in the diff could not be found in the
                repository.

            reviewboard.scmtools.core.SCMError:
                There was an error talking to the repository when validating
                the existence of a file.

            reviewboard.scmtools.git.ShortSHA1Error:
                A SHA1 specified in the diff was in the short form, which
                could not be used to look up the file. This is applicable only
                to Git.
        """
        check_diff_size(diff_file, parent_diff_file)

        if parent_diff_file:
            parent_diff_file_name = parent_diff_file.name
            parent_diff_file_contents = parent_diff_file.read()
        else:
            parent_diff_file_name = None
            parent_diff_file_contents = None

        return self.create_from_data(
            repository=repository,
            diff_file_name=diff_file.name,
            diff_file_contents=diff_file.read(),
            parent_diff_file_name=parent_diff_file_name,
            parent_diff_file_contents=parent_diff_file_contents,
            request=request,
            validate_only=validate_only,
            **kwargs)


class DiffCommitManager(BaseDiffManager):
    """A custom manager for DiffCommit objects.

    This includes utilities for creating diffsets based on the data from form
    uploads, webapi requests, and upstream repositories.
    """

    def by_diffset_ids(
        self,
        diffset_ids: Sequence[int],
    ) -> Mapping[int, Sequence[DiffCommit]]:
        """Return the commits grouped by DiffSet IDs.

        Args:
            diffset_ids (list of int):
                The primary keys of the DiffSets to retrieve commits for.

        Returns:
            dict:
            A mapping of :py:class:`~reviewboard.diffviewer.models.diffset.
            DiffSet` primary keys to lists of
            :py:class:`DiffCommits <reviewboard.diffviewer.models.diffcommit.
            DiffCommit>`.
        """
        commits_by_diffset_id = {
            pk: []
            for pk in diffset_ids
        }

        if diffset_ids:
            for commit in self.filter(diffset_id__in=diffset_ids):
                commits_by_diffset_id[commit.diffset_id].append(commit)

        return commits_by_diffset_id

    def create_from_data(self,
                         repository,
                         diff_file_name,
                         diff_file_contents,
                         parent_diff_file_name,
                         parent_diff_file_contents,
                         diffset,
                         commit_id,
                         parent_id,
                         commit_message,
                         author_name,
                         author_email,
                         author_date,
                         validation_info=None,
                         request=None,
                         committer_name=None,
                         committer_email=None,
                         committer_date=None,
                         base_commit_id=None,
                         check_existence=True,
                         validate_only=False):
        """Create a DiffCommit from raw diff data.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the diff was posted against.

            diff_file_name (unicode):
                The name of the diff file.

            diff_file_contents (bytes):
                The contents of the diff file.

            parent_diff_file_name (unicode):
                The name of the parent diff file.

            parent_diff_file_contents (bytes):
                The contents of the parent diff file.

            diffset (reviewboard.diffviewer.models.diffset.DiffSet):
                The DiffSet to attach the created DiffCommit to.

            commit_id (unicode):
                The unique identifier of the commit.

            parent_id (unicode):
                The unique identifier of the parent commit.

            commit_message (unicode):
                The commit message.

            author_name (unicode):
                The author's name.

            author_email (unicode):
                The author's e-mail address.

            author_date (datetime.datetime):
                The date and time that the commit was authored.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            validation_info (dict, optional):
                A dictionary of parsed validation information from the
                :py:class:`~reviewboard.webapi.resources.validate_diffcommit.
                ValidateDiffCommitResource`.

            committer_name (unicode, optional):
                The committer's name.

            committer_email (unicode, optional)
                The committer's e-mail address.

            committer_date (datetime.datetime, optional):
                The date and time that the commit was committed.

            base_commit_id (unicode, optional):
                The ID of the commit that the diff is based upon. This is
                needed by some SCMs or hosting services to properly look up
                files, if the diffs represent blob IDs instead of commit IDs
                and the service doesn't support those lookups.

            check_existence (bool, optional):
                Whether or not existence checks should be performed against
                the upstream repository.

            validate_only (bool, optional):
                Whether to just validate and not save. If ``True``, then this
                won't populate the database at all and will return ``None``
                upon success. This defaults to ``False``.

        Returns:
            reviewboard.diffviewer.models.diffcommit.DiffCommit:
            The created model instance.
        """
        diffcommit = self.model(
            filename=diff_file_name,
            diffset=diffset,
            commit_id=commit_id,
            parent_id=parent_id,
            author_name=author_name,
            author_email=author_email,
            author_date=author_date,
            commit_message=commit_message,
            committer_name=committer_name,
            committer_email=committer_email,
            committer_date=committer_date)

        if not validate_only:
            diffcommit.save()

        get_file_exists = partial(get_file_exists_in_history,
                                  validation_info=validation_info or {},
                                  repository=repository,
                                  parent_id=parent_id)

        create_filediffs(
            get_file_exists=get_file_exists,
            diff_file_contents=diff_file_contents,
            parent_diff_file_contents=parent_diff_file_contents,
            repository=repository,
            request=request,
            basedir='',
            base_commit_id=base_commit_id,
            diffset=diffset,
            diffcommit=diffcommit,
            validate_only=validate_only,
            check_existence=check_existence)

        if validate_only:
            return None

        return diffcommit


class DiffSetManager(BaseDiffManager):
    """A custom manager for DiffSet objects.

    This includes utilities for creating diffsets based on the data from form
    uploads, webapi requests, and upstream repositories.
    """

    def create_from_data(self,
                         repository,
                         diff_file_name,
                         diff_file_contents,
                         parent_diff_file_name=None,
                         parent_diff_file_contents=None,
                         diffset_history=None,
                         basedir=None,
                         request=None,
                         base_commit_id=None,
                         check_existence=True,
                         validate_only=False,
                         **kwargs):
        """Create a DiffSet from raw diff data.

        This parses a diff and optional parent diff covering one or more files,
        validates, and constructs :py:class:`DiffSets
        <reviewboard.diffviewer.models.diffset.DiffSet>` and
        :py:class:`FileDiffs <reviewboard.diffviewer.models.filediff.FileDiff>`
        representing the diff.

        This can optionally validate the diff without saving anything to the
        database. In this case, no value will be returned. Instead, callers
        should take any result as success.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the diff applies to.

            diff_file_name (unicode):
                The filename of the main diff file.

            diff_file_contents (bytes):
                The contents of the main diff file.

            parent_diff_file_name (unicode, optional):
                The filename of the parent diff, if one is provided.

            parent_diff_file_contents (bytes, optional):
                The contents of the parent diff, if one is provided.

            diffset_history (reviewboard.diffviewer.models.diffset_history.
                             DiffSetHistory, optional):
                The history object to associate the DiffSet with. This is
                not required if using ``validate_only=True``.

            basedir (unicode, optional):
                The base directory to prepend to all file paths in the diff.

            request (django.http.HttpRequest, optional):
                The current HTTP request, if any. This will result in better
                logging.

            base_commit_id (unicode, optional):
                The ID of the commit that the diff is based upon. This is
                needed by some SCMs or hosting services to properly look up
                files, if the diffs represent blob IDs instead of commit IDs
                and the service doesn't support those lookups.

            check_existence (bool, optional):
                Whether to check for file existence as part of the validation
                process. This defaults to ``True``.

            validate_only (bool, optional):
                Whether to just validate and not save. If ``True``, then this
                won't populate the database at all and will return ``None``
                upon success. This defaults to ``False``.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The resulting DiffSet stored in the database, if processing
            succeeded and ``validate_only=False``.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the main diff or parent diff.

            reviewboard.diffviewer.errors.EmptyDiffError:
                The provided diff file did not contain any file changes.

            reviewboard.scmtools.core.FileNotFoundError:
                A file specified in the diff could not be found in the
                repository.

            reviewboard.scmtools.core.SCMError:
                There was an error talking to the repository when validating
                the existence of a file.

            reviewboard.scmtools.git.ShortSHA1Error:
                A SHA1 specified in the diff was in the short form, which
                could not be used to look up the file. This is applicable only
                to Git.
        """
        diffset = self.model(
            name=diff_file_name,
            revision=0,
            basedir=basedir,
            history=diffset_history,
            repository=repository,
            diffcompat=DiffCompatVersion.DEFAULT,
            base_commit_id=base_commit_id)

        if not validate_only:
            diffset.save()

        create_filediffs(
            get_file_exists=repository.get_file_exists,
            diff_file_contents=diff_file_contents,
            parent_diff_file_contents=parent_diff_file_contents,
            repository=repository,
            request=request,
            basedir=basedir,
            base_commit_id=base_commit_id,
            diffset=diffset,
            check_existence=check_existence,
            validate_only=validate_only
        )

        if validate_only:
            return None

        return diffset

    def create_empty(self, repository, diffset_history=None, **kwargs):
        """Create a DiffSet with no attached FileDiffs.

        An empty DiffSet must be created before
        :py:class:`DiffCommits <reviewboard.diffviewer.models.diffcommit.
        DiffCommit>` can be added to it. When the DiffCommits are created,
        the :py:class:`FileDiffs <reviewboard.diffviewer.models.filediff.
        FileDiff>` will be attached to the DiffSet.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to create the DiffSet in.

            diffset_history (reviewboard.diffviewer.models.diffset_history.
                             DiffSetHistory, optional):
                An optional DiffSetHistory instance to attach the DiffSet to.

            **kwargs (dict):
                Additional keyword arguments to pass to :py:meth:`create`.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The created DiffSet.
        """
        kwargs.setdefault('revision', 0)
        return super(DiffSetManager, self).create(
            name='diff',
            history=diffset_history,
            repository=repository,
            diffcompat=DiffCompatVersion.DEFAULT,
            **kwargs)



"""Unit tests for the SSO backend registry."""

from django.http import HttpResponse
from django.urls import NoReverseMatch, path, reverse
from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError

from reviewboard.accounts.sso.backends import sso_backends
from reviewboard.accounts.sso.backends.base import BaseSSOBackend
from reviewboard.testing import TestCase


def backend_test_view(request, backend_id):
    return HttpResponse(str(backend_id))


class SSOBackendRegistryTests(TestCase):
    """Unit tests for the SSO backend registry."""

    class DummyBackend(BaseSSOBackend):
        backend_id = 'dummy'
        name = 'Dummy'

    class DummyBackendWithURLs(BaseSSOBackend):
        backend_id = 'dummy-with-urls'
        name = 'DummyWithURLs'

        @property
        def urls(self):
            return [
                path('sso-endpoint/',
                     backend_test_view,
                     name='dummy-backend-sso-endpoint')
            ]

    def setUp(self):
        """Set up the test case."""
        self.dummy_backend = self.DummyBackend()
        self.dummy_backend_with_urls = self.DummyBackendWithURLs()

    def tearDown(self):
        """Tear down the test case."""
        super(SSOBackendRegistryTests, self).tearDown()

        try:
            sso_backends.unregister(self.dummy_backend)
        except ItemLookupError:
            pass

        try:
            sso_backends.unregister(self.dummy_backend_with_urls)
        except ItemLookupError:
            pass

    def test_register_without_urls(self):
        """Testing SSO backend registration"""
        sso_backends.register(self.dummy_backend)

        with self.assertRaises(AlreadyRegisteredError):
            sso_backends.register(self.dummy_backend)

    def test_unregister_without_urls(self):
        """Testing SSO backend unregistration"""
        sso_backends.register(self.dummy_backend)
        sso_backends.unregister(self.dummy_backend)

    def test_register_with_urls(self):
        """Testing SSO backend registration with URLs"""
        sso_backends.register(self.dummy_backend_with_urls)

        self.assertEqual(
            reverse(
                'sso:dummy-with-urls:dummy-backend-sso-endpoint',
                kwargs={
                    'backend_id': 'dummy-with-urls',
                }),
            '/account/sso/dummy-with-urls/sso-endpoint/')

        with self.assertRaises(AlreadyRegisteredError):
            sso_backends.register(self.dummy_backend_with_urls)

    def test_unregister_with_urls(self):
        """Testing SSO backend registration with URLs"""
        sso_backends.register(self.dummy_backend_with_urls)
        sso_backends.unregister(self.dummy_backend_with_urls)

        with self.assertRaises(NoReverseMatch):
            reverse(
                'sso:dummy-with-urls:dummy-backend-sso-endpoint',
                kwargs={
                    'backend_id': 'dummy-with-urls',
                })



"""The base search backend classes."""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from djblets.siteconfig.models import SiteConfiguration
from haystack import connections
from haystack.utils.loading import load_backend


class SearchBackendForm(forms.Form):
    """A search backend configuration form.

    This form allows the configuration of a Haystack backend to be configured
    from the admin interface and reloaded on the fly.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the backend.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.
        """
        self.request = kwargs.pop('request')

        super(SearchBackendForm, self).__init__(*args, **kwargs)


class SearchBackend(object):
    """A search backend.

    A SearchBackend is a wrapper around a Haystack backend that has methods for
    getting and setting per-backend configuration. This allows the backend to
    be configured at runtime and live reloaded, instead of having them
    hardcoded in :file:`settings.py`.
    """

    #: The search engine's unique identifier.
    search_backend_id = None

    #: The human-readable name for the search engine.
    name = None

    #: The name of the Haystack search engine backend.
    haystack_backend_name = None

    #: The configuration form class for the search engine.
    config_form_class = None

    #: The default search engine settings.
    default_settings = {}

    #: A mapping of search engine settings to form fields.
    form_field_map = {}

    @property
    def configuration(self):
        """The configuration for the search engine.

        Returns:
            dict:
            The configuration for the search engine.
        """
        engine_settings = (
            SiteConfiguration.objects
            .get_current()
            .get('search_backend_settings')
            .get(self.search_backend_id, {})
        )

        configuration = {
            key: engine_settings.get(key, self.default_settings[key])
            for key in self.default_settings.keys()
        }
        configuration['ENGINE'] = self.haystack_backend_name

        return configuration

    @configuration.setter
    def configuration(self, value):
        """Set the configuration for the search engine.

        This does not save the configuration to the database.
        :py:meth:`djblets.siteconfig.models.SiteConfiguration.save` must be
        called.

        Args:
            value (dict):
                The configuration to set.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        search_backend_settings = siteconfig.get('search_backend_settings')
        engine_settings = \
            search_backend_settings.setdefault(self.search_backend_id, {})

        engine_settings.update({
            key: value[key]
            for key in self.default_settings.keys()
            if key in value
        })

        siteconfig.set('search_backend_settings', search_backend_settings)
        siteconfig.save(update_fields=('settings',))

    def get_configuration_from_form_data(self, form_data):
        """Return the configuration from the form's data.

         Args:
             form_data (dict):
                The form data.

        Returns:
            dict:
            The search engine configuration.
        """
        return {
            config_key: form_data.get(field_name)
            for field_name, config_key in self.form_field_map.items()
        }

    def get_form_data(self):
        """Return the form data for the current configuration.

        Returns:
            dict:
            The search engine form data.
        """
        configuration = self.configuration

        return {
            field_name: configuration[config_key]
            for field_name, config_key in self.form_field_map.items()
        }

    def load_haystack_engine(self, **kwargs):
        """Load the Haystack backend engine.

        This will construct the engine with the provided parameters. A new
        instance will be constructed every time this is called.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to the engine constructor.

        Returns:
            django_haystack.backends.BaseEngine:
            The resulting engine instance.
        """
        engine_cls = load_backend(self.haystack_backend_name)

        return engine_cls(**kwargs)

    def validate(self, configuration, **kwargs):
        """Validate the settings and dependencies for the search backend.

        This will perform a test search to ensure the configuration works.

        Subclasses can override this to perform other checks that may be
        required (such as checking for a suitable module). They should call
        the parent method in order to perform a test search using the
        configured settings.

        Version Changed:
            4.0.5:
            Added the ``configuration`` and ``**kwargs`` arguments. Subclasses
            msut be updated to take ``**kwargs``.

        Raises:
            django.core.exceptions.ValidationError:
                Backend configuration, supporting modules, or the test
                search failed.
        """
        new_config = configuration.copy()
        new_config['SILENTLY_FAIL'] = False

        using = 'validation-test'
        connections.connections_info[using] = dict({
            'ENGINE': self.haystack_backend_name,
        }, **new_config)
        connections.reload(using)

        engine = self.load_haystack_engine(using=using)
        backend = engine.backend(using, **new_config)

        try:
            backend.search('___reviewboard-search-config-test___')
        except Exception as e:
            raise ValidationError(
                _('Performing a test query failed. Make sure your '
                  'configuration is correct. The error we received from the '
                  'search backend was: %s')
                % e)
        finally:
            connections.connections_info[using] = {}

    def get_config_form(self, data=None, **kwargs):
        """Create and return a new configuration form instance.

        The returned form will have a prefix of the search engine ID.

        Args:
            data (dict, optional):
                The form data.

            **kwargs (dict):
                Additional keyword arguments to pass to the form.

        Returns:
            SearchBackendForm:
            The instantiated form class.
        """
        return self.config_form_class(initial=self.get_form_data(),
                                      data=data,
                                      prefix=self.search_backend_id,
                                      **kwargs)



import json
import logging

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html_join
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_protect
from djblets.cache.forwarding_backend import DEFAULT_FORWARD_CACHE_ALIAS
from djblets.siteconfig.views import site_settings as djblets_site_settings

from reviewboard.admin.cache_stats import get_cache_stats
from reviewboard.admin.decorators import superuser_required
from reviewboard.admin.forms.ssh_settings import SSHSettingsForm
from reviewboard.admin.security_checks import SecurityCheckRunner
from reviewboard.admin.support import get_support_url, serialize_support_data
from reviewboard.admin.widgets import (admin_widgets_registry,
                                       dynamic_activity_data)
from reviewboard.ssh.client import SSHClient
from reviewboard.ssh.utils import humanize_key


logger = logging.getLogger(__name__)


@staff_member_required
def admin_dashboard_view(request):
    """Display the administration dashboard.

    This is the entry point to the admin site, containing news updates and
    useful administration tasks.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        django.http.HttpResponse:
        The resulting HTTP response for the view.
    """
    widgets_info = []
    widgets_html = []

    for widget_cls in admin_widgets_registry:
        try:
            widget = widget_cls()

            if not widget.can_render(request):
                continue

            if widget.dom_id is None:
                widget.dom_id = 'admin-widget-%s' % widget.widget_id

            widget_info = {
                'id': widget.widget_id,
                'domID': widget.dom_id,
                'viewClass': widget.js_view_class,
                'modelClass': widget.js_model_class,
            }

            js_view_options = widget.get_js_view_options(request)
            js_model_attrs = widget.get_js_model_attrs(request)
            js_model_options = widget.get_js_model_options(request)

            if js_view_options:
                widget_info['viewOptions'] = js_view_options

            if js_model_attrs:
                widget_info['modelAttrs'] = js_model_attrs

            if js_model_options:
                widget_info['modelOptions'] = js_model_options

            widget_html = widget.render(request)
        except Exception as e:
            logger.exception('Error setting up administration widget %r: %s',
                             widget_cls, e, extra={'request': request})
            continue

        widgets_info.append(widget_info)
        widgets_html.append((widget_html,))

    return render(
        request=request,
        template_name='admin/dashboard.html',
        context={
            'page_model_attrs': {
                'supportData': serialize_support_data(request,
                                                      force_is_admin=True),
                'widgetsData': widgets_info,
            },
            'title': _('Admin Dashboard'),
            'widgets_html': format_html_join('', '{0}', widgets_html),
        })


@staff_member_required
def cache_stats(request, template_name="admin/cache_stats.html"):
    """Display statistics on the cache.

    This includes such pieces of information as memory used, cache misses, and
    uptime.
    """
    cache_stats = get_cache_stats()
    cache_info = settings.CACHES[DEFAULT_FORWARD_CACHE_ALIAS]

    return render(
        request=request,
        template_name=template_name,
        context={
            'cache_hosts': cache_stats,
            'cache_backend': cache_info['BACKEND'],
            'title': _('Server Cache'),
            'root_path': reverse('admin:index'),
        })


@staff_member_required
def security(request, template_name="admin/security.html"):
    """Run security checks and report the results."""
    runner = SecurityCheckRunner()
    results = runner.run()

    return render(
        request=request,
        template_name=template_name,
        context={
            'test_results': results,
            'title': _('Security Checklist'),
        })


@superuser_required
def site_settings(request, form_class, template_name='admin/settings.html'):
    """Render the general site settings page."""
    return djblets_site_settings(request, form_class, template_name, {
        'root_path': reverse('admin:index'),
    })


@csrf_protect
@superuser_required
def ssh_settings(request, template_name='admin/ssh_settings.html'):
    """Render the SSH settings page."""
    client = SSHClient()
    key = client.get_user_key()

    if request.method == 'POST':
        form = SSHSettingsForm(request.POST, request.FILES)

        if form.is_valid():
            if form.did_request_delete() and client.get_user_key() is not None:
                try:
                    form.delete()
                    return HttpResponseRedirect('.')
                except Exception as e:
                    logger.error('Deleting SSH key failed: %s', e,
                                 extra={'request': request})
            else:
                try:
                    form.create(request.FILES)
                    return HttpResponseRedirect('.')
                except Exception as e:
                    # Fall through. It will be reported inline and in the log.
                    logger.error('Uploading SSH key failed: %s', e,
                                 extra={'request': request})
    else:
        form = SSHSettingsForm()

    if key:
        fingerprint = humanize_key(key)
    else:
        fingerprint = None

    return render(
        request=request,
        template_name=template_name,
        context={
            'has_file_field': True,
            'key': key,
            'fingerprint': fingerprint,
            'public_key': client.get_public_key(key).replace('\n', ''),
            'form': form,
        })


def manual_updates_required(request, updates):
    """Render a page showing required updates that the admin must make.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        updates (list):
            The list of required updates to display on the page.

    Returns:
        django.http.HttpResponse:
        The response to send to the client.
    """
    return render(
        request=request,
        template_name='admin/manual_updates_required.html',
        context={
            'updates': [
                render_to_string(template_name=update_template_name,
                                 context=extra_context,
                                 request=request)
                for update_template_name, extra_context in updates
            ],
        })


def widget_activity(request):
    """Return JSON data for the admin activity widget."""
    activity_data = dynamic_activity_data(request)

    return HttpResponse(json.dumps(activity_data),
                        content_type="application/json")


def support_redirect(request, **kwargs):
    """Return an HttpResponseRedirect to the Beanbag support page."""
    return HttpResponseRedirect(get_support_url(request))



"""Unit tests for reviewboard.diffviewer.settings.DiffSettings.

Version Added:
    5.0.2
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from reviewboard.diffviewer.settings import DiffSettings
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from typelets.json import JSONDict


class DiffSettingsTests(TestCase):
    """Unit tests for reviewboard.diffviewer.settings.DiffSettings."""

    def test_create(self) -> None:
        """Testing DiffSettings.create"""
        siteconfig_settings: JSONDict = {
            'code_safety_checkers': {
                'trojan_code': {
                    'enable_confusables': True,
                },
            },
            'diffviewer_context_num_lines': 10,
            'diffviewer_custom_pygments_lexers': {
                '.foo': 'SomeLexer',
            },
            'diffviewer_include_space_patterns': ['*.a', '*.b'],
            'diffviewer_paginate_by': 20,
            'diffviewer_paginate_orphans': 5,
            'diffviewer_syntax_highlighting': True,
            'diffviewer_syntax_highlighting_threshold': 10_000,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create()

        self.assertEqual(diff_settings.code_safety_configs, {
            'trojan_code': {
                'enable_confusables': True,
            },
        })
        self.assertEqual(diff_settings.context_num_lines, 10)
        self.assertEqual(diff_settings.custom_pygments_lexers, {
            '.foo': 'SomeLexer',
        })
        self.assertEqual(diff_settings.include_space_patterns,
                         ['*.a', '*.b'])
        self.assertEqual(diff_settings.paginate_by, 20)
        self.assertEqual(diff_settings.paginate_orphans, 5)
        self.assertTrue(diff_settings.interdiff_filtering)
        self.assertTrue(diff_settings.syntax_highlighting)
        self.assertEqual(diff_settings.syntax_highlighting_threshold,
                         10_000)

    def test_create_with_siteconfig_syntax_highlighting_true(self) -> None:
        """Testing DiffSettings.create with
        siteconfig.diffviewer_syntax_highlighting=True
        """
        siteconfig_settings: JSONDict = {
            'diffviewer_syntax_highlighting': True,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create()

        self.assertTrue(diff_settings.syntax_highlighting)

    def test_create_with_siteconfig_syntax_highlighting_false(self) -> None:
        """Testing DiffSettings.create with
        siteconfig.diffviewer_syntax_highlighting=False
        """
        siteconfig_settings: JSONDict = {
            'diffviewer_syntax_highlighting': False
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create()

        self.assertFalse(diff_settings.syntax_highlighting)

    def test_create_with_siteconfig_syntax_highlighting_true_user_true(
        self,
    ) -> None:
        """Testing DiffSettings.create with
        siteconfig.diffviewer_syntax_highlighting=True and
        Profile.syntax_highlighting=True
        """
        user = self.create_user()
        profile = user.get_profile()  # type: ignore
        profile.syntax_highlighting = True

        siteconfig_settings: JSONDict = {
            'diffviewer_syntax_highlighting': True,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create(user=user)

        self.assertTrue(diff_settings.syntax_highlighting)

    def test_create_with_siteconfig_syntax_highlighting_true_user_false(
        self,
    ) -> None:
        """Testing DiffSettings.create with
        siteconfig.diffviewer_syntax_highlighting=True and
        Profile.syntax_highlighting=False
        """
        user = self.create_user()
        profile = user.get_profile()  # type: ignore
        profile.syntax_highlighting = False

        siteconfig_settings: JSONDict = {
            'diffviewer_syntax_highlighting': True,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create(user=user)

        self.assertFalse(diff_settings.syntax_highlighting)

    def test_create_with_local_site_layer(self) -> None:
        """Testing DiffSettings.create with a local site that has settings
        overrides
        """
        local_site = self.create_local_site(extra_data={
            'siteconfig': {
                'diffviewer_context_num_lines': 5,
            },
        })
        siteconfig_settings: JSONDict = {
            'diffviewer_context_num_lines': 10,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create(local_site=local_site)

        self.assertEqual(diff_settings.context_num_lines, 5)

    def test_create_with_local_site_no_settings(self) -> None:
        """Testing DiffSettings.create with a local site that does not have
        settings overrides
        """
        local_site = self.create_local_site(extra_data={})
        siteconfig_settings: JSONDict = {
            'diffviewer_context_num_lines': 10,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create(local_site=local_site)

        self.assertEqual(diff_settings.context_num_lines, 10)

    def test_tab_size(self) -> None:
        """Testing DiffSettings.tab_size normalization"""
        with self.siteconfig_settings(
            {'diffviewer_default_tab_size': None},
        ):
            diff_settings = DiffSettings.create()
            self.assertEqual(diff_settings.tab_size,
                             DiffSettings.DEFAULT_TAB_SIZE)

        with self.siteconfig_settings(
            {'diffviewer_default_tab_size': 0},
        ):
            diff_settings = DiffSettings.create()
            self.assertEqual(diff_settings.tab_size,
                             DiffSettings.DEFAULT_TAB_SIZE)

        with self.siteconfig_settings(
            {'diffviewer_default_tab_size': 4},
        ):
            diff_settings = DiffSettings.create()
            self.assertEqual(diff_settings.tab_size, 4)

    def test_state_hash(self) -> None:
        """Testing DiffSettings.state_hash"""
        siteconfig_settings: JSONDict = {
            'code_safety_checkers': {
                'trojan_code': {
                    'enable_confusables': True,
                },
            },
            'diffviewer_context_num_lines': 10,
            'diffviewer_custom_pygments_lexers': {
                '.foo': 'SomeLexer',
            },
            'diffviewer_default_tab_size': 4,
            'diffviewer_include_space_patterns': ['*.a', '*.b'],
            'diffviewer_paginate_by': 20,
            'diffviewer_paginate_orphans': 5,
            'diffviewer_syntax_highlighting': True,
            'diffviewer_syntax_highlighting_threshold': 10_000,
        }

        with self.siteconfig_settings(siteconfig_settings):
            diff_settings = DiffSettings.create()

        self.assertEqual(
            diff_settings.state_hash,
            '7723b408c3a75c3be994f20b946ed34945831490ea1f1d6d703e197140923a28')



"""Unit tests for reviewboard.accounts.forms.pages.ThemeForm.

Version Added:
    7.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import kgb
from django.contrib import messages

from reviewboard.accounts.forms.pages import ThemeForm
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.http import HttpRequest

    from reviewboard.accounts.models import Profile


class ThemeFormTests(kgb.SpyAgency, TestCase):
    """Unit tests for ThemeForm.

    Version Added:
        7.0
    """

    ######################
    # Instance variables #
    ######################

    request: HttpRequest
    profile: Profile
    user: User

    def setUp(self) -> None:
        super(ThemeFormTests, self).setUp()

        self.request = self.create_http_request()
        self.user = self.create_user()
        self.profile = self.user.get_profile()

        # Suppresses MessageFailure Exception at the end of save()
        self.spy_on(messages.add_message)

    def test_init(self) -> None:
        """Testing ThemeForm.__init__"""
        form = ThemeForm(page=None,
                         request=self.request,
                         user=self.user)

        self.assertEqual(form.fields['ui_theme'].choices, [
            ('default', 'Default appearance (System theme)'),
            ('light', 'Light mode'),
            ('dark', 'Dark mode'),
        ])

    def test_load_with_defaults(self) -> None:
        """Testing ThemeForm.load with profile defaults"""
        form = ThemeForm(page=None,
                         request=self.request,
                         user=self.user)

        form.load()

        self.assertEqual(form['ui_theme'].initial, 'default')

    def test_load_with_ui_theme(self) -> None:
        """Testing ThemeForm.load with ui_theme"""
        self.profile.ui_theme_id = 'system'

        form = ThemeForm(page=None,
                         request=self.request,
                         user=self.user)
        form.load()

        self.assertEqual(form['ui_theme'].initial, 'system')

    def test_save(self) -> None:
        """Testing ThemeForm.save"""
        form = ThemeForm(
            page=None,
            request=self.request,
            user=self.user,
            data={
                'ui_theme': 'dark',
            })
        form.is_valid()
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(self.profile.ui_theme_id, 'dark')

        self.assertSpyCalledWith(
            messages.add_message,
            self.request,
            messages.INFO,
            'Your appearance settings have been saved.')



"""Unit tests for reviewboard.datagrids.columns.SummaryColumn.

Version Added:
    5.0.7
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.accounts.models import ReviewRequestVisit
from reviewboard.datagrids.columns import SummaryColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase
from reviewboard.reviews.models import ReviewRequest


class SummaryColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.SummaryColumn."""

    column = SummaryColumn()

    def test_render_data(self):
        """Testing SummaryColumn.render_data"""
        review_request = self.create_review_request(summary='Summary 1',
                                                    publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '<span>Summary 1</span>')

    def test_render_data_with_draft(self):
        """Testing SummaryColumn.render_data with draft review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-draft">Draft</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_draft_summary(self):
        """Testing SummaryColumn.render_data with draft summary"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = 'Draft Summary 1'
        review_request.visibility = ReviewRequestVisit.VISIBLE

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-draft">Draft</label>'
                         '<span>Draft Summary 1</span>')

    def test_render_data_with_draft_and_no_summary(self):
        """Testing SummaryColumn.render_data with draft and no summary"""
        review_request = self.create_review_request(
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        review_request.summary = None

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-draft">Draft</label>'
                         '<span class="no-summary">No Summary</span>')

    def test_render_data_with_archived(self):
        """Testing SummaryColumn.render_data with archived review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user,
            publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.ARCHIVED

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-archived">Archived</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_muted(self):
        """Testing SummaryColumn.render_data with muted review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user,
            publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.MUTED

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-muted">Muted</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_draft_and_archived(self):
        """Testing SummaryColumn.render_data with draft and archived
        review request
        """
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.ARCHIVED

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-draft">Draft</label>'
                         '<label class="label-archived">Archived</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_draft_and_muted(self):
        """Testing SummaryColumn.render_data with draft and muted
        review request
        """
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.MUTED

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-draft">Draft</label>'
                         '<label class="label-muted">Muted</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_submitted(self):
        """Testing SummaryColumn.render_data with submitted review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            status=ReviewRequest.SUBMITTED,
            submitter=self.request.user,
            public=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-submitted">Completed</label>'
                         '<span>Summary 1</span>')

    def test_render_data_with_discarded(self):
        """Testing SummaryColumn.render_data with discarded review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            status=ReviewRequest.DISCARDED,
            submitter=self.request.user,
            public=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value,
                         '<label class="label-discarded">Discarded</label>'
                         '<span>Summary 1</span>')

    def test_to_json(self) -> None:
        """Testing SummaryColumn.to_json"""
        review_request = self.create_review_request(summary='Summary 1',
                                                    publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [],
                'summary': 'Summary 1',
            })

    def test_to_json_with_draft(self) -> None:
        """Testing SummaryColumn.to_json with draft review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Draft',
                        'status': 'draft',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_draft_summary(self) -> None:
        """Testing SummaryColumn.to_json with draft summary"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = 'Draft Summary 1'
        review_request.visibility = ReviewRequestVisit.VISIBLE

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Draft',
                        'status': 'draft',
                    },
                ],
                'summary': 'Draft Summary 1',
            })

    def test_to_json_with_draft_and_no_summary(self) -> None:
        """Testing SummaryColumn.to_json with draft and no summary"""
        review_request = self.create_review_request(
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        review_request.summary = None

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Draft',
                        'status': 'draft',
                    },
                ],
                'summary': None,
            })

    def test_to_json_with_archived(self) -> None:
        """Testing SummaryColumn.to_json with archived review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user,
            publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.ARCHIVED

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Archived',
                        'status': 'archived',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_muted(self) -> None:
        """Testing SummaryColumn.to_json with muted review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user,
            publish=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.MUTED

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Muted',
                        'status': 'muted',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_draft_and_archived(self) -> None:
        """Testing SummaryColumn.to_json with draft and archived
        review request
        """
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.ARCHIVED

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Draft',
                        'status': 'draft',
                    },
                    {
                        'label': 'Archived',
                        'status': 'archived',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_draft_and_muted(self) -> None:
        """Testing SummaryColumn.to_json with draft and muted
        review request
        """
        review_request = self.create_review_request(
            summary='Summary 1',
            submitter=self.request.user)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.MUTED

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Draft',
                        'status': 'draft',
                    },
                    {
                        'label': 'Muted',
                        'status': 'muted',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_submitted(self) -> None:
        """Testing SummaryColumn.to_json with submitted review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            status=ReviewRequest.SUBMITTED,
            submitter=self.request.user,
            public=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Completed',
                        'status': 'completed',
                    },
                ],
                'summary': 'Summary 1',
            })

    def test_to_json_with_discarded(self) -> None:
        """Testing SummaryColumn.to_json with discarded review request"""
        review_request = self.create_review_request(
            summary='Summary 1',
            status=ReviewRequest.DISCARDED,
            submitter=self.request.user,
            public=True)

        # These are generally set by the column's augment_queryset().
        review_request.draft_summary = None
        review_request.visibility = ReviewRequestVisit.VISIBLE

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            {
                'labels': [
                    {
                        'label': 'Discarded',
                        'status': 'discarded',
                    },
                ],
                'summary': 'Summary 1',
            })



"""Add Review.general_comments field.

Version Added:
    3.0
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Review', 'general_comments', models.ManyToManyField,
             related_model='reviews.GeneralComment'),
]



"""Administration form for e-mail settings."""

import logging

from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.sites.models import Site
from django.utils.translation import gettext, gettext_lazy as _
from djblets.siteconfig.forms import SiteSettingsForm
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.siteconfig import load_site_config
from reviewboard.notifications.email.message import EmailMessage


logger = logging.getLogger(__name__)


class EMailSettingsForm(SiteSettingsForm):
    """E-mail settings for Review Board."""

    FROM_SPOOFING_CHOICES = (
        (EmailMessage.FROM_SPOOFING_SMART,
         _('Auto (when allowed by DMARC record)')),
        (EmailMessage.FROM_SPOOFING_ALWAYS, _('Always')),
        (EmailMessage.FROM_SPOOFING_NEVER, _('Never')),
    )

    mail_send_review_mail = forms.BooleanField(
        label=_('Send e-mails for review requests and reviews'),
        required=False)
    mail_send_review_close_mail = forms.BooleanField(
        label=_('Send e-mails when review requests are closed'),
        required=False)
    mail_send_new_user_mail = forms.BooleanField(
        label=_('Send e-mails to administrators when new users register '
                'accounts'),
        required=False)
    mail_send_password_changed_mail = forms.BooleanField(
        label=_('Send e-mails to users when they change their password'),
        required=False)
    mail_enable_autogenerated_header = forms.BooleanField(
        label=_('Enable "Auto-Submitted: auto-generated" header'),
        help_text=_('Marks outgoing e-mails as "auto-generated" to avoid '
                    'auto-replies. Disable this if your mailing list rejects '
                    '"auto-generated" e-mails.'),
        required=False)
    mail_default_from = forms.CharField(
        label=_('Default From address'),
        help_text=_('The e-mail address that all e-mails will be sent from. '
                    'The "Sender" header will be used to make e-mails appear '
                    'to come from the user triggering the e-mail.'),
        required=False,
        widget=forms.TextInput(attrs={'size': '50'}))
    mail_from_spoofing = forms.ChoiceField(
        label=_("Use the user's From address"),
        choices=FROM_SPOOFING_CHOICES,
        required=False,
        help_text=_('Depending on the sender domain\'s DKIM/SPF '
                    'configuration, e-mails may be marked as suspicious when '
                    'sending from the user\'s own e-mail address. '
                    '<strong>Auto</strong> will check the domain\'s DMARC '
                    'record before using their address, falling back to the '
                    '"Default From address" above. <strong>Always</strong> '
                    'will use it unconditionally. <strong>Never</strong> will '
                    'use the default address for every e-mail.'))
    mail_host = forms.CharField(
        label=_('Mail server'),
        required=False,
        widget=forms.TextInput(attrs={'size': '50'}))
    mail_port = forms.IntegerField(
        label=_('Port'),
        required=False,
        widget=forms.TextInput(attrs={'size': '5'}))
    mail_host_user = forms.CharField(
        label=_('Username'),
        required=False,
        widget=forms.TextInput(attrs={'size': '30', 'autocomplete': 'off'}))
    mail_host_password = forms.CharField(
        widget=forms.PasswordInput(attrs={'size': '30', 'autocomplete': 'off'},
                                   render_value=True),
        label=_('Password'),
        required=False)
    mail_use_tls = forms.BooleanField(
        label=_('Use TLS for authentication'),
        required=False)

    send_test_mail = forms.BooleanField(
        label=_('Send a test e-mail after saving'),
        help_text=_('Send an e-mail to yourself using these server settings.'),
        required=False)

    def clean_mail_host(self):
        """Clean the mail_host field.

        This will strip all whitespace around the value.

        Returns:
            unicode:
            The ``mail_host`` field, with all whitespace stripped.
        """
        return self.cleaned_data['mail_host'].strip()

    def save(self):
        """Save the form.

        This will write the new configuration to the database. It will then
        force a site configuration reload.
        """
        super(EMailSettingsForm, self).save()

        # Reload any important changes into the Django settings.
        load_site_config()

        if self.cleaned_data['send_test_mail']:
            site = Site.objects.get_current()
            siteconfig = SiteConfiguration.objects.get_current()
            product_name = settings.PRODUCT_NAME
            request = self.request

            site_url = '%s://%s' % (siteconfig.get('site_domain_method'),
                                    site.domain)

            if request and request.user.is_authenticated:
                to_user = request.user.email
            else:
                to_user = siteconfig.get('site_admin_email')

            try:
                message = EmailMessage(
                    subject=(
                        gettext('%s e-mail settings test')
                        % product_name
                    ),
                    text_body=(
                        gettext("This is a test of the e-mail settings "
                                "for the %(product)s server at %(url)s. "
                                "If you got this, you're all set!")
                        % {
                            'product': product_name,
                            'url': site_url,
                        }
                    ),
                    from_email=siteconfig.get('mail_default_from'),
                    to=[to_user])

                message.send()
            except Exception as e:
                error = str(e)

                if request is not None:
                    messages.error(
                        request,
                        gettext('Failed to send the test e-mail: "%s". Check '
                                'the server logs for additional details.')
                        % error)

                logger.exception('Failed to send test e-mail to %s: %s',
                                 to_user, error)

    class Meta:
        title = _('E-Mail Settings')
        save_blacklist = ('send_test_mail',)

        fieldsets = (
            {
                'title': _('E-Mail Notification Settings'),
                'classes': ('wide',),
                'fields': ('mail_send_review_mail',
                           'mail_send_review_close_mail',
                           'mail_send_new_user_mail',
                           'mail_send_password_changed_mail'),
            },
            {
                'title': _('E-Mail Delivery Settings'),
                'classes': ('wide',),
                'fields': ('mail_default_from',
                           'mail_from_spoofing',
                           'mail_enable_autogenerated_header'),
            },
            {
                'title': _('E-Mail Server Settings'),
                'classes': ('wide',),
                'fields': ('mail_host', 'mail_port', 'mail_host_user',
                           'mail_host_password', 'mail_use_tls',
                           'send_test_mail'),
            },
        )



"""Hosting service for Bugzilla."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService
from reviewboard.admin.validation import validate_bug_tracker_base_hosting_url

if TYPE_CHECKING:
    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class BugzillaForm(BaseHostingServiceRepositoryForm):
    bugzilla_url = forms.CharField(
        label=_('Bugzilla URL'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        validators=[validate_bug_tracker_base_hosting_url])

    def clean_bugzilla_url(self):
        return self.cleaned_data['bugzilla_url'].rstrip('/')


class Bugzilla(BaseHostingService, BaseBugTracker):
    """Hosting service for Bugzilla."""

    name = 'Bugzilla'
    hosting_service_id = 'bugzilla'
    form = BugzillaForm
    bug_tracker_field = '%(bugzilla_url)s/show_bug.cgi?id=%%s'
    supports_bug_trackers = True

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.base.bug_tracker.BugInfo:
            Information about the bug.
        """
        # This requires making two HTTP requests: one for the summary and
        # status, and one to get the "first comment" (description).
        bug_id = str(bug_id)

        result: BugInfo = {
            'summary': '',
            'description': '',
            'status': '',
        }

        bugzilla_url = repository.extra_data['bug_tracker-bugzilla_url']
        url = f'{bugzilla_url}/rest/bug/{bug_id}'

        try:
            rsp = self.client.json_get(
                f'{url}?include_fields=summary,status')[0]
            result['summary'] = rsp['bugs'][0]['summary']
            result['status'] = rsp['bugs'][0]['status']
        except Exception as e:
            logger.warning('Unable to fetch bugzilla data from %s: %s',
                           url, e, exc_info=True)

        try:
            url += '/comment'
            rsp = self.client.json_get(url)[0]
            result['description'] = rsp['bugs'][bug_id]['comments'][0]['text']
        except Exception as e:
            logger.warning('Unable to fetch bugzilla data from %s: %s',
                           url, e, exc_info=True)

        return result



"""Administration form for avatar settings."""

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext, gettext_lazy as _
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.avatars import avatar_services


class AvatarServicesForm(SiteSettingsForm):
    """A form for managing avatar services."""

    avatars_enabled = forms.BooleanField(
        label=_('Enable avatars'),
        help_text=_('We recommend enabling avatars on all installations, '
                    'and enabling or disabling the services you prefer.'),
        required=False)

    enabled_services = forms.MultipleChoiceField(
        label=_('Enabled services'),
        required=False,
        help_text=_("If no services are enabled, a fallback avatar will be "
                    "shown that displays the user's initials."),
        widget=forms.CheckboxSelectMultiple())

    default_service = forms.ChoiceField(
        label=_('Default service'),
        help_text=_('The avatar service to be used by default for users who '
                    'do not have an avatar service configured. This must be '
                    'one of the enabled avatar services below.'),
        required=False
    )

    def __init__(self, *args, **kwargs):
        """Initialize the settings form.

        This will populate the choices and initial values for the form fields
        based on the current avatar configuration.

        Args:
            *args (tuple):
                Additional positional arguments for the parent class.

            **kwargs (dict):
                Additional keyword arguments for the parent class.
        """
        super(AvatarServicesForm, self).__init__(*args, **kwargs)

        default_choices = [('none', 'None')]
        enable_choices = []

        for service in avatar_services:
            default_choices.append((service.avatar_service_id, service.name))
            enable_choices.append((service.avatar_service_id, service.name))

        default_service_field = self.fields['default_service']
        enabled_services_field = self.fields['enabled_services']

        default_service_field.choices = default_choices
        enabled_services_field.choices = enable_choices
        enabled_services_field.initial = [
            service.avatar_service_id
            for service in avatar_services.enabled_services
        ]

        default_service = avatar_services.default_service

        if avatar_services.default_service is not None:
            default_service_field.initial = default_service.avatar_service_id

    def clean_enabled_services(self):
        """Clean and validate the enabled_services field.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if an unknown service is attempted to be enabled.
        """
        for service_id in self.cleaned_data['enabled_services']:
            if not avatar_services.has_service(service_id):
                raise ValidationError(
                    gettext('"%s" is not an available avatar service.')
                    % service_id)

        return self.cleaned_data['enabled_services']

    def clean(self):
        """Clean and validate the form.

        This will clean the form, handling any fields that need cleaned
        that depend on the cleaned data of other fields.

        Raises:
            django.core.exceptions.ValidationError:
                Raised if an unknown service or disabled service is set to be
                the default.
        """
        cleaned_data = super(AvatarServicesForm, self).clean()

        enabled_services = set(cleaned_data['enabled_services'])
        service_id = cleaned_data['default_service']

        if service_id == 'none':
            default_service = None
        else:
            if not avatar_services.has_service(service_id):
                raise ValidationError(
                    gettext('"%s" is not an available avatar service.')
                    % service_id)
            elif service_id not in enabled_services:
                raise ValidationError(
                    gettext('The "%s" avatar service is disabled and cannot '
                            'be set.')
                    % service_id)

            default_service = avatar_services.get('avatar_service_id',
                                                  service_id)

        cleaned_data['default_service'] = default_service

        return cleaned_data

    def save(self):
        """Save the enabled services and default service to the database."""
        avatar_services.set_enabled_services(
            [
                avatar_services.get('avatar_service_id', service_id)
                for service_id in self.cleaned_data['enabled_services']
            ],
            save=False)

        avatar_services.set_default_service(
            self.cleaned_data['default_service'],
            save=False)

        avatar_services.avatars_enabled = self.cleaned_data['avatars_enabled']
        avatar_services.save()

    class Meta:
        title = _('Avatar Settings')
        fieldsets = (
            {
                'fields': (
                    'avatars_enabled',
                    'enabled_services',
                    'default_service',
                ),
            },
        )



from datetime import datetime, timedelta, timezone

from django.utils import timezone as django_timezone
from djblets.db.query import get_object_or_none
from djblets.webapi.errors import (INVALID_FORM_DATA,
                                   PERMISSION_DENIED)
from djblets.webapi.testing.decorators import webapi_test_template
from kgb import SpyAgency

from reviewboard.site.models import LocalSite
from reviewboard.webapi.errors import TOKEN_GENERATION_FAILED
from reviewboard.webapi.models import WebAPIToken
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (api_token_item_mimetype,
                                                api_token_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_api_token_item_url,
                                           get_api_token_list_url)


def _compare_item(self, item_rsp, api_token):
    if api_token.last_used:
        expected_last_used = api_token.last_used.isoformat()
    else:
        expected_last_used = None

    if api_token.expires:
        expected_expires = api_token.expires.isoformat()
    else:
        expected_expires = None

    if api_token.invalid_date:
        expected_invalid_date = api_token.invalid_date.isoformat()
    else:
        expected_invalid_date = None

    self.assertEqual(item_rsp['id'], api_token.pk)
    self.assertEqual(item_rsp['token'], api_token.token)
    self.assertEqual(item_rsp['token_generator_id'],
                     api_token.token_generator_id)
    self.assertEqual(item_rsp['note'], api_token.note)
    self.assertEqual(item_rsp['policy'], api_token.policy)
    self.assertEqual(item_rsp['extra_data'], api_token.extra_data)
    self.assertEqual(item_rsp['time_added'], api_token.time_added.isoformat())
    self.assertEqual(item_rsp['last_updated'],
                     api_token.last_updated.isoformat())
    self.assertEqual(item_rsp['last_used'], expected_last_used)
    self.assertEqual(item_rsp['expires'], expected_expires)
    self.assertEqual(item_rsp['valid'], api_token.valid)
    self.assertEqual(item_rsp['invalid_date'], expected_invalid_date)
    self.assertEqual(item_rsp['invalid_reason'], api_token.invalid_reason)


class APITokenTestsMixin(object):
    token_data = {
        'note': 'This is my new token.',
        'policy': '{"perms": "ro", "resources": {"*": {"allow": ["*"]}}}',
    }


class ResourceListTests(SpyAgency, ExtraDataListMixin, BaseWebAPITestCase,
                        APITokenTestsMixin, metaclass=BasicTestsMetaclass):
    """Testing the APITokenResource list APIs."""
    fixtures = ['test_users']
    sample_api_url = 'users/<username>/api-tokens/'
    resource = resources.api_token
    test_api_token_access = False
    test_oauth_token_access = False

    compare_item = _compare_item

    def setup_http_not_allowed_item_test(self, user):
        return get_api_token_list_url(user)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        if populate_items:
            if not with_local_site:
                LocalSite.objects.create(name=self.local_site_name)

            # Due to running this test with a WebAPIToken, we may
            # already have one in the database we need to include.
            items = list(user.webapi_tokens.all())
            items.append(self.create_webapi_token(
                user, note='Token 1', with_local_site=with_local_site))

            self.create_webapi_token(user, note='Token 2',
                                     with_local_site=not with_local_site)
        else:
            items = []

        return (get_api_token_list_url(user, local_site_name),
                api_token_list_mimetype,
                items)

    @webapi_test_template
    def test_get_with_api_token_auth_denied(self) -> None:
        """Testing the GET <URL> API denies access when using token-based
        authentication
        """
        setup_state = self.setup_get_list_test_state(with_webapi_token=True)

        rsp = self.api_get(setup_state['url'], expected_status=403)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        if post_valid_data:
            post_data = self.token_data.copy()
        else:
            post_data = {}

        return (get_api_token_list_url(user, local_site_name),
                api_token_item_mimetype,
                post_data,
                [local_site_name])

    def check_post_result(self, user, rsp, local_site_name):
        token_rsp = rsp['api_token']
        token = WebAPIToken.objects.get(pk=token_rsp['id'])

        self.compare_item(token_rsp, token)

        if local_site_name:
            self.assertIsNotNone(token.local_site_id)
            self.assertEqual(token.local_site.name, local_site_name)
        else:
            self.assertIsNone(token.local_site_id)

    @webapi_test_template
    def test_post_with_generation_error(self):
        """Testing the POST <URL> API with Token Generation Failed error"""
        def _generate_token(self, *args, **kwargs):
            kwargs['max_attempts'] = 0
            orig_generate_token(*args, **kwargs)

        orig_generate_token = WebAPIToken.objects.generate_token

        self.spy_on(WebAPIToken.objects.generate_token,
                    call_fake=_generate_token)

        rsp = self.api_post(get_api_token_list_url(self.user),
                            self.token_data,
                            expected_status=500)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], TOKEN_GENERATION_FAILED.code)
        self.assertEqual(rsp['err']['msg'],
                         'Could not create a unique API token. '
                         'Please try again.')

    @webapi_test_template
    def test_post_with_api_token_auth_denied(self) -> None:
        """Testing the POST <URL> API denies access when using token-based
        authentication
        """
        setup_state = self.setup_post_test_state(with_webapi_token=True)

        rsp = self.api_post(setup_state['url'], expected_status=403)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    @webapi_test_template
    def test_post_with_expires_field_utc_timezone(self):
        """Testing the POST <URL> API with setting an ISO 8601
        timestamp with UTC timezone for the expiration date
        """
        expires = datetime(2022, 9, 20, 13, 42, 0,
                           tzinfo=timezone.utc).isoformat()
        token_data = self.token_data.copy()
        token_data['expires'] = expires

        rsp = self.api_post(get_api_token_list_url(self.user),
                            token_data,
                            expected_mimetype=api_token_item_mimetype,
                            expected_status=201)

        token = WebAPIToken.objects.get(pk=rsp['api_token']['id'])

        self.assertEqual(token.expires.isoformat(),
                         '2022-09-20T13:42:00+00:00')
        self.check_post_result(self.user, rsp, None)

    @webapi_test_template
    def test_post_with_expires_field_non_utc_timezone(self):
        """Testing the POST <URL> API with setting an ISO 8601
        timestamp with a non-UTC timezone for the expiration date
        """
        expires = datetime(
            2022, 9, 20, 13, 42, 0,
            tzinfo=django_timezone.get_fixed_timezone(timedelta(hours=5)))
        utc_expires = '2022-09-20T08:42:00+00:00'

        token_data = self.token_data.copy()
        token_data['expires'] = expires.isoformat()

        rsp = self.api_post(get_api_token_list_url(self.user),
                            token_data,
                            expected_mimetype=api_token_item_mimetype,
                            expected_status=201)

        # Compare against the UTC time since dates are stored in UTC.
        rsp['api_token']['expires'] = utc_expires

        token = WebAPIToken.objects.get(pk=rsp['api_token']['id'])

        self.assertEqual(token.expires.isoformat(), utc_expires)
        self.check_post_result(self.user, rsp, None)

    @webapi_test_template
    def test_post_with_expires_field_no_timezone(self):
        """Testing the POST <URL> API with setting an ISO 8601
        timestamp with no timezone specified for the expiration date
        """
        # When no timezone is specified, the user's profile's timezone
        # will be used.
        profile = self.user.get_profile()
        profile.timezone = 'America/New_York'
        profile.save(update_fields=('timezone',))

        expires = datetime(2022, 9, 20, 13, 42, 0)
        utc_expires = '2022-09-20T17:42:00+00:00'

        token_data = self.token_data.copy()
        token_data['expires'] = expires.isoformat()

        rsp = self.api_post(get_api_token_list_url(self.user),
                            token_data,
                            expected_mimetype=api_token_item_mimetype,
                            expected_status=201)

        # Compare against the UTC time since dates are stored in UTC.
        rsp['api_token']['expires'] = utc_expires

        token = WebAPIToken.objects.get(pk=rsp['api_token']['id'])

        self.assertEqual(token.expires.isoformat(), utc_expires)
        self.check_post_result(self.user, rsp, None)

    @webapi_test_template
    def test_post_with_expires_field_empty(self):
        """Testing the POST <URL> API with setting an empty expiration
        date for the token
        """
        token_data = self.token_data.copy()
        token_data['expires'] = ''

        rsp = self.api_post(get_api_token_list_url(self.user),
                            token_data,
                            expected_mimetype=api_token_item_mimetype,
                            expected_status=201)

        self.check_post_result(self.user, rsp, None)


class ResourceItemTests(ExtraDataItemMixin, BaseWebAPITestCase,
                        APITokenTestsMixin, metaclass=BasicTestsMetaclass):
    """Testing the APITokenResource item APIs."""
    fixtures = ['test_users']
    sample_api_url = 'users/<username>/api-tokens/<id>/'
    resource = resources.api_token
    test_api_token_access = False
    test_oauth_token_access = False

    compare_item = _compare_item

    def setup_http_not_allowed_list_test(self, user):
        token = self.create_webapi_token(user)

        return get_api_token_item_url(token)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        token = self.create_webapi_token(user, with_local_site=with_local_site)

        return (get_api_token_item_url(token, local_site_name),
                [token.pk])

    def check_delete_result(self, user, token_id):
        self.assertIsNone(get_object_or_none(WebAPIToken, pk=token_id))

    @webapi_test_template
    def test_delete_with_api_token_auth_denied(self) -> None:
        """Testing the DELETE <URL> API denies access when using token-based
        authentication
        """
        setup_state = self.setup_delete_test_state(with_webapi_token=True)

        rsp = self.api_delete(setup_state['url'], expected_status=403)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        token = self.create_webapi_token(user, with_local_site=with_local_site)

        return (get_api_token_item_url(token, local_site_name),
                api_token_item_mimetype,
                token)

    @webapi_test_template
    def test_get_not_modified(self):
        """Testing the GET <URL> API with Not Modified response"""
        token = self.create_webapi_token(self.user)

        self._testHttpCaching(get_api_token_item_url(token),
                              check_last_modified=True)

    @webapi_test_template
    def test_get_with_api_token_auth_denied(self) -> None:
        """Testing the GET <URL> API denies access when using token-based
        authentication
        """
        setup_state = self.setup_get_item_test_state(with_webapi_token=True)

        rsp = self.api_get(setup_state['url'], expected_status=403)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    #
    # HTTP PUT tests
    #

    @webapi_test_template
    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        token = self.create_webapi_token(user, with_local_site=with_local_site)

        return (get_api_token_item_url(token, local_site_name),
                api_token_item_mimetype,
                self.token_data.copy(),
                token,
                [])

    def check_put_result(self, user, item_rsp, token):
        self.compare_item(item_rsp, WebAPIToken.objects.get(pk=token.pk))

    @webapi_test_template
    def test_put_with_api_token_auth_denied(self):
        """Testing the PUT <URL> API denies access when using
        token-based authentication
        """
        setup_state = self.setup_put_test_state(with_webapi_token=True)

        rsp = self.api_put(setup_state['url'], expected_status=403)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    @webapi_test_template
    def test_put_with_invalidation(self):
        """Testing the PUT <URL> API with invalidating a token"""
        token = self.create_webapi_token(self.user)
        url = get_api_token_item_url(token)

        # Testing that the token can be invalidated.
        rsp = self.api_put(
            url,
            {
                'valid': 0,
                'invalid_reason': 'test reason',
            },
            expected_mimetype=api_token_item_mimetype,
            expected_status=200)

        self.check_put_result(self.user, rsp['api_token'], token)

        # Testing that the token cannot be updated to be valid.
        rsp = self.api_put(url,
                           {'valid': 1},
                           expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)

    @webapi_test_template
    def test_put_with_expires_field_utc_timezone(self):
        """Testing the PUT <URL> API with setting an ISO 8601
        timestamp with UTC timezone for the expiration date
        """
        token = self.create_webapi_token(self.user)
        expires = datetime(2022, 9, 20, 13, 42, 0,
                           tzinfo=timezone.utc).isoformat()

        token_data = self.token_data.copy()
        token_data['expires'] = expires

        rsp = self.api_put(get_api_token_item_url(token),
                           token_data,
                           expected_mimetype=api_token_item_mimetype,
                           expected_status=200)

        token.refresh_from_db()

        self.assertEqual(token.expires.isoformat(),
                         '2022-09-20T13:42:00+00:00')
        self.check_put_result(self.user, rsp['api_token'], token)

    @webapi_test_template
    def test_put_with_expires_field_non_utc_timezone(self):
        """Testing the PUT <URL> API with setting an ISO 8601
        timestamp with a non-UTC timezone for the expiration date
        """
        token = self.create_webapi_token(self.user)

        expires = datetime(
            2022, 9, 20, 13, 42, 0,
            tzinfo=django_timezone.get_fixed_timezone(timedelta(hours=5)))
        utc_expires = '2022-09-20T08:42:00+00:00'

        token_data = self.token_data.copy()
        token_data['expires'] = expires.isoformat()

        rsp = self.api_put(get_api_token_item_url(token),
                           token_data,
                           expected_mimetype=api_token_item_mimetype,
                           expected_status=200)

        # Compare against the UTC time since dates are stored in UTC.
        rsp['api_token']['expires'] = utc_expires

        token.refresh_from_db()

        self.assertEqual(token.expires.isoformat(), utc_expires)
        self.check_put_result(self.user, rsp['api_token'], token)

    @webapi_test_template
    def test_put_with_expires_field_no_timezone(self):
        """Testing the PUT <URL> API with setting an ISO 8601
        timestamp with no timezone specified for the expiration date
        """
        # When no timezone is specified, the user's profile's timezone
        # will be used.
        profile = self.user.get_profile()
        profile.timezone = 'America/New_York'
        profile.save(update_fields=('timezone',))

        token = self.create_webapi_token(self.user)

        expires = datetime(2022, 9, 20, 13, 42, 0)
        utc_expires = '2022-09-20T17:42:00+00:00'

        token_data = self.token_data.copy()
        token_data['expires'] = expires.isoformat()

        rsp = self.api_put(get_api_token_item_url(token),
                           token_data,
                           expected_mimetype=api_token_item_mimetype,
                           expected_status=200)

        # Compare against the UTC time since dates are stored in UTC.
        rsp['api_token']['expires'] = utc_expires

        token.refresh_from_db()

        self.assertEqual(token.expires.isoformat(), utc_expires)
        self.check_put_result(self.user, rsp['api_token'], token)

    @webapi_test_template
    def test_put_with_expires_field_empty(self):
        """Testing the PUT <URL> API with setting an empty expiration
        date for the token
        """
        token = self.create_webapi_token(self.user)
        url = get_api_token_item_url(token)

        token_data = self.token_data.copy()
        token_data['expires'] = ''

        rsp = self.api_put(url,
                           token_data,
                           expected_mimetype=api_token_item_mimetype,
                           expected_status=200)

        token.refresh_from_db()

        self.assertIsNone(token.expires)
        self.check_put_result(self.user, rsp['api_token'], token)

    @webapi_test_template
    def test_put_with_expires_resets_notification_flag(self):
        """Testing the PUT <URL> API with setting an expiration resets
        the expired notification sent flag on the token
        """
        token = self.create_webapi_token(
            self.user,
            expires=datetime(2022, 9, 20, 13, 42, 0, tzinfo=timezone.utc),
            extra_data={'expired_notification_sent': True})

        token_data = self.token_data.copy()
        token_data['expires'] = ''

        rsp = self.api_put(get_api_token_item_url(token),
                           token_data,
                           expected_mimetype=api_token_item_mimetype,
                           expected_status=200)

        token.refresh_from_db()

        self.assertIsNone(token.expires)
        self.assertFalse(token.extra_data['expired_notification_sent'])
        self.check_put_result(self.user, rsp['api_token'], token)
